/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* This is the marcel threading library for OMPi
 */
 
#ifndef __EE_H__
#define __EE_H__

#include "../../compiler/config.h"

#ifdef HAVE_SPINLOCKS
  /* Needed to get spin locks, before including <pthread.h> */
  #define _XOPEN_SOURCE 600
#endif

#include <unistd.h>
#include <marcel.h>
#include "../sysdeps.h"

/* The threading library should provide the following types:
 *    othr_key_t
 *    othr_lock_t
 */
 
typedef struct {            /* For nested locks */
  marcel_mutex_t lock;     /* The lock in question */
  marcel_mutex_t ilock;    /* Lock to access the whole struct */
  marcel_cond_t  cond;     /* For waiting until lock is available */
  int count;                /* # times locked by the same thread */
  marcel_t owner;          /* Which thread owns the nestable lock */
} othr_nestlock_t;

typedef union {
  struct {
    int   type;                 /* normal/spin/nested */
    union {
      marcel_mutex_t normal;   /* normal lock */
      othr_nestlock_t nest;      /* nest lock */
      struct {
        int             rndelay;     /* Used for initial spin delays */
        marcel_mutex_t mutex;
      } spin;
    } data;
  } lock;

  char padding[CACHE_LINE];
} othr_lock_t;


/* 
 * The threading library should provide the functions that follow.
 */

/* Base functions */
extern int  othr_initialize(int *argc, char ***argv,
                            ort_icvs_t *icv, ort_caps_t *cap);
extern void othr_finalize(int exitvalue);
extern int  othr_request(int numthreads, int level);
extern void othr_create(int numthreads, int level, void *arg, void **info);
extern void othr_waitall(void **info);
#define othr_yield() marcel_yield()

/* Thread-specifics */
#define othr_key_t            marcel_key_t
#define othr_key_create(a,b)  marcel_key_create(a,b)
#define othr_getspecific(a)   marcel_getspecific(a)
#define othr_setspecific(a,b) marcel_setspecific(a,b)
  
/* Locks */
extern int  othr_init_lock(othr_lock_t *lock, int type);
extern int  othr_destroy_lock(othr_lock_t *lock);
extern int  othr_set_lock(othr_lock_t *lock);
extern int  othr_unset_lock(othr_lock_t *lock);
extern int  othr_test_lock(othr_lock_t *lock);

#endif
